/*
 * parse.c
 * The taxman's taken all my dough.
 *
 * Created by mercadal on Fri Feb 01 2002.
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmfgetln.h"
#include "common.h"


#define NUMDEFTAGS 16


static void defaultHandler(char *);
static void linkHandler(char *);
static void printLinkTitle(void);
static void readUdefs(FILE *);
static void titleHandler(char *);


void initializeTags(FILE *, char);
void parseTag(void);
void uninitializeTags(void);


struct ftnt {
    char *there;
    char *repl;
};

typedef struct {
    struct ftnt *myTag;
    void (*handler)(char *);
} mtag;


static char *title;
static char *linkloc;
static char quiet;
static int numtags = NUMDEFTAGS;	/* default set has NUMDEFTAGS mtags */
static int numudefs = 0;
static mtag *knowTags;


static struct ftnt deftags[NUMDEFTAGS] = {
    {"title", ""},	/* keep at position 0 */
    {"link", ""},	/* keep at position 1 */
    {"channel", "<b>"},
    {"/channel", "</b><br><hr>"},
    {"item", ""},
    {"/item", "<br>"},
    {"/title", ""},
    {"/link", ""},
    {"description", "<br>"},
    {"/description", "<br>"},
    {"copyright", "<!--RSS input file copyright:  "},
    {"/copyright", " -->"},
    {"language", "<!-- "},
    {"/language"," -->"},
    {"*RDFSTART*", ""},	/* keep second from the last */
    {"*RDFEND*", ""}	/* keep the last */
};

static struct ftnt *udefs;


/*
 * Let's initialize these mtags!
 */
void initializeTags(FILE *udeffile, char runquiet)
{
    register int cnt;
    
    
    /* should we run quiet? */
    quiet = runquiet;
    
    /* 
     * Our user didn't specify their own user-defined tags to 
     * run in the parser.  As such, we'll use the default 
     * rss2html set. 
     */
    knowTags = (mtag *)xmalloc(numtags * sizeof(mtag));

    for (cnt = 0; cnt < numtags; cnt++)
        switch (cnt) {
            case 0: /* title case, need title handler */
                knowTags[cnt].myTag = &deftags[cnt];
                knowTags[cnt].handler = titleHandler;
                break;
            case 1: /* link case */
                knowTags[cnt].myTag = &deftags[cnt];
                knowTags[cnt].handler = linkHandler;
                break;
            default:
                knowTags[cnt].myTag = &deftags[cnt];
                knowTags[cnt].handler = defaultHandler;
                break;
        }
        
    if (udeffile != NULL)
        readUdefs(udeffile);
    
    /* Let's do the *RDFSTART* stuff */
    knowTags[NUMDEFTAGS - 2].handler(knowTags[NUMDEFTAGS - 2].myTag->repl);
}


/*
 * The parser for the so-called user defined tags file 
 * is pretty poor and could probably do with some gussying
 * up by the more intrepid coder.  Namely, we realloc for each 
 * line which is not necessarily efficient.
 */
void readUdefs(FILE *from)
{
    char *line;
    register char reassigned;
    register int cnt, index;
    

    udefs = (struct ftnt *)xmalloc(1 * sizeof(struct ftnt));
    cnt = 0;
    
    /* 
     * Can I do this more efficiently? 
     * Say 16 bit Intel assembly and I kill you.  :-)
     */
    while ((line = cmfgetln(from)) != NULL) {
        if ((isspace(*line)) || (*line == '\0') || (*line == '#'))
            continue;
        
        udefs[cnt].there = strdup(strtok(line, "\\"));
        udefs[cnt].repl = strdup(strtok(NULL, "\\"));
    
        cnt++;
        udefs = (struct ftnt *)xrealloc(udefs, (cnt + 1) * sizeof(struct ftnt));
        free(line);
    }
    
    numudefs = cnt;
    
    for (index = 0, reassigned = 0; index < numudefs; index++, reassigned = 0) {
        for (cnt = 0; cnt < numtags; cnt++) {
            if (strcmp(knowTags[cnt].myTag->there, udefs[index].there) == NULL) {
                knowTags[cnt].myTag = &udefs[index];
                reassigned = 1;
                break;
            }
        }
        
        if (!reassigned) {
            numtags++;
            knowTags = (mtag *)xrealloc(knowTags, numtags * sizeof(mtag));
            knowTags[numtags - 1].myTag = &udefs[index];
            knowTags[numtags - 1].handler = defaultHandler;
        }
    }
}


void uninitializeTags(void)
{
    register int cnt;
    
    
    /* Do *RDFEND* stuff */
    knowTags[NUMDEFTAGS - 1].handler(knowTags[NUMDEFTAGS - 1].myTag->repl);
    
    for (cnt = 0; cnt < numudefs; cnt++) {
        free(udefs[cnt].there);
        free(udefs[cnt].repl);
    }
    
    free(udefs);
    free(knowTags);
}


void parseTag(void)
{
    char *buf;
    register int cnt;


    buf = cmfgettoch(stdin, '>');
    
    for (cnt = 0; cnt < numtags; cnt++) {
        if (strcmp(knowTags[cnt].myTag->there, buf) == NULL) {
            free(buf);
            
            /* Run this tag's handler */
            knowTags[cnt].handler(knowTags[cnt].myTag->repl);
            return;
        }
    }
    
    /* Unknown tag */
    if (!quiet)
        (void)fprintf(stderr, "Don't know %s tag\n", buf);
    
    free(buf);
}


void titleHandler(char *print)
{
    char *minbuf;


    defaultHandler(print);

    minbuf = cmfgettoch(stdin, '<');
    title = strdup(minbuf);
    free(minbuf);
    
    if (linkloc != NULL)
        printLinkTitle();
    
    parseTag(); /* parse the closing tag that's left */
}


void linkHandler(char *print)
{
    char *minbuf;


    defaultHandler(print);

    minbuf = cmfgettoch(stdin, '<');
    linkloc = strdup(minbuf);
    free(minbuf);
    
    if (title != NULL)
        printLinkTitle();
    
    parseTag(); /* parse close tag left */
}


void printLinkTitle(void)
{

    (void)printf("<a href=\"%s\">%s</a>\n", linkloc, title);
    free(linkloc);
    free(title);
    
    /* Reset, linkHandler & titleHandler assume a href is printed */
    linkloc = title = NULL;
}


void defaultHandler(char *print)
{

    (void)printf("%s", print);
}
